(* ::Package:: *)

BeginPackage["TenReS`"]


Global`TenReS::usage="A few definitions need to be made by the user. Type ?CoeffQ, ?Specialization, and ?CyclicModules for more information.";


CoeffQ::usage="TenReS`CoeffQ[k_] should check whether k is in the ring over which the tensor product is formed.";


Specialization::usage="TenReS`Specialization should store a list of the form {Rule[y1,X[y1]],Rule[y2,X[y2]],...} specifying the refinement of the alphabet.";


CyclicModules::usage="TenReS`CyclicModules should store a list of pairs {x,gen} where the module corresponding to the letter x is generated by the element gen.";


Prod::usage="Prod[m1,m2,...] represents the tensor product.";


ExtractReducibleWords::usage="ExtractReducibleWords[sys] gives the list of words corresponding to the rules in the reduction system sys.";


ApplyRules::usage="ApplyRules[expr,sys] reduces expr by the reduction system sys.";


IrreducibleWords::usage="IrreducibleWords[alphabet,reducible,n] generates all irreducible words of length at most n.";


ReducibleWordQ::usage="ReducibleWordQ[w,red] checks whether the word w contains a subword that is (a specialization of) a word in the list red of reducible words.";


Overlap::usage="Overlap[ABC,C,A] designates an overlap ambiguity of the words AB and BC.";
Inclusion::usage="Inclusion[ABC,A,C] designates an inclusion ambiguity of the words B and ABC.";
SpecialOverlap::usage="SpecialOverlap[ABC,C,A] designates an overlap ambiguity of two words \!\(\*SubscriptBox[\(AB\), \(1\)]\) and \!\(\*SubscriptBox[\(B\), \(2\)]\)C, where B is a common specialization of \!\(\*SubscriptBox[\(B\), \(1\)]\) and \!\(\*SubscriptBox[\(B\), \(2\)]\).";
SpecialInclusion::usage="SpecialInclusion[ABC,A,C] designates an inclusion ambiguity of two words \!\(\*SubscriptBox[\(B\), \(1\)]\) and \!\(\*SubscriptBox[\(AB\), \(2\)]\)C, where B is a common specialization of \!\(\*SubscriptBox[\(B\), \(1\)]\) and \!\(\*SubscriptBox[\(B\), \(2\)]\).";


GenerateOverlaps::usage="Generate overlap ambiguities for a given list of reducible words.";
GenerateInclusions::usage="Generate inclusion ambiguities for a given list of reducible words.";
GenerateSpecialOverlaps::usage="Generate overlap ambiguities with specialization for a given list of reducible words.";
GenerateSpecialInclusions::usage="Generate inclusion ambiguities with specialization for a given list of reducible words.";
GenerateAmbiguities::usage="Generate all ambiguities for a given list of reducible words.";


SPoly::usage="For an ambiguity amb of the reduction system sys, SPoly[amb,sys] generates the corresponding S-polynomial.";


CheckResolvability::usage="CheckResolvability[sys] returns a list of all ambiguities that could not be resolved together with the reduced form of their S-polynomial.";


Begin["`Private`"]


(* ::Subsubsection:: *)
(*Data types*)


ReductionSystem={RepeatedNull[{_List,_}]};


(* ::Subsubsection:: *)
(*Tensor algebra*)


CoeffQ[_Integer]:=True


Prod[a___,0,b___]=0;
Prod[a___,b_Plus,c___]:=(Prod[a,#,c]&/@b)
(*Prod[a___,d_Integer*b_,c___]:=d Prod[a,b,c]*)
Prod[a___,Prod[b___],c___]:=Prod[a,b,c]


Prod[a___,c_?(CoeffQ[#]&&(#=!=1)&),b___]:=c Prod[a,1,b]
Prod[a___,(c_?CoeffQ)f_,b___]:=c Prod[a,f,b]


(* ::Subsubsection:: *)
(*Reduction system*)


(* ::Text:: *)
(*This implementation assumes that a leading term occurs at most once among the rules and that the leading terms of any two rules cannot be specialized to a common term.*)


(* ::Text:: *)
(*Consequently FindRule[word,sys] is supposed to find at most one matching rule.*)


FindRule[word_List,sys:ReductionSystem]:=
Select[sys,(word===#[[1]]||Specialize[word,#[[1]]]=!={{},{}})&]


(* ::Text:: *)
(*Specialize[v,w] returns a list of two lists of rules {rv,rw} such that ReplacePart[v,rv]=ReplacePart[w,rw], if a common specialization of v and w exists, or rv=rw={} otherwise.*)


Specialize[v_List,w_List]:=
If[Length[v]=!=Length[w],{{},{}},
	Module[{i,rv={},rw={}},
		For[i=1,i<=Length[v],i++,Which[
			v[[i]]===w[[i]],
				Null,
			MemberQ[v[[i]]/.Specialization,w[[i]]],
				AppendTo[rv,i->w[[i]]],
			MemberQ[w[[i]]/.Specialization,v[[i]]],
				AppendTo[rw,i->v[[i]]],
			True,
				rv={};rw={};i=Length[v]
		]];
		{rv,rw}
	]
]


ExtractRules[sys:ReductionSystem]:=
ExtractRules[sys]=Module[{a,b,h,i,j,m,p},
	a=Unique[];b=Unique[];
	m=Table[Unique[],{Max[1,Length/@First/@sys]}];
	Table[
		p=Table[PatternTest[Pattern[Evaluate[m[[j]]],Blank[]],Evaluate[Subscript[MemberQ,sys[[i,1,j]]]]],{j,Length[sys[[i,1]]]}];
		p=Join[{Pattern[Evaluate[a],BlankNullSequence[]]},p,{Pattern[Evaluate[b],BlankNullSequence[]]}];
		(RuleDelayed@@{Prod@@p,Prod[a,h@@Take[m,Length[sys[[i,1]]]],b]})/.h->sys[[i,2]]
	,{i,Length[sys]}]
]


ExtractRules[sys:ReductionSystem,Print->True]:=
ExtractRules[sys,Print->True]=Module[{a,b,h1,h2,h3,i,j,m,p},
	a=Unique[];b=Unique[];
	m=Table[Unique[],{Max[1,Length/@First/@sys]}];
	Table[
		p=Table[PatternTest[Pattern[Evaluate[m[[j]]],Blank[]],Evaluate[Subscript[MemberQ,sys[[i,1,j]]]]],{j,Length[sys[[i,1]]]}];
		p=Join[{Pattern[Evaluate[a],BlankNullSequence[]]},p,{Pattern[Evaluate[b],BlankNullSequence[]]}];
		(RuleDelayed@@{Prod@@p,h1@@{h2[sys[[i]]],Prod[a,h3@@Take[m,Length[sys[[i,1]]]],b]}})/.{h1->CompoundExpression,h2->Print,h3->sys[[i,2]]}
	,{i,Length[sys]}]
]


ExtractReducibleWords[sys:ReductionSystem]:=sys[[All,1]]


ApplyRules[expr_,sys:ReductionSystem,OptionsPattern[{Print->False,Count->False}]]:=
Module[{n,tmp,p,c},
	p=(OptionValue[Print]===True);
	c=(OptionValue[Count]===True);
	If[c,
		tmp=FixedPointList[ReplaceAll[If[p,ExtractRules[sys,Print->True],ExtractRules[sys]]],expr];
		Print[Length[tmp]-2," steps"];
		Last[tmp]
	,
		expr//.If[p,ExtractRules[sys,Print->True],ExtractRules[sys]]
	]
]


IrreducibleWords[reduciblewords_List,level:Alternatives[_Integer?Positive,{_Integer?Positive}]]:=
Module[{X,Z},
	Z=DeleteDuplicates[Join[Sequence@@reduciblewords,Sequence@@Specialization[[All,2]],Specialization[[All,1]]]];
	X=DeleteDuplicates[Replace[Z,ReplacePart[Specialization,{_,2,0}->Sequence],{1}]];
	IrreducibleWords[X,Replace[reduciblewords,ReplacePart[Specialization,{_,2,0}->Alternatives],{2}],level]
]


IrreducibleWords[alphabet_List,reduciblewords_List,level:Alternatives[_Integer?Positive,{_Integer?Positive}]]:=
Module[{n,w,X,irr1,irr2,red},
	If[MatchQ[level,_Integer],n=level,n=level[[1]]];
	X=DeleteDuplicates[Replace[alphabet,ReplacePart[Specialization,{_,2,0}->Sequence],{1}]];
	red=Replace[reduciblewords,ReplacePart[Specialization,{_,2,0}->Alternatives],{2}];
	If[n===1,
		DeleteCases[{#}&/@X,Alternatives@@red]
	,
		irr1=IrreducibleWords[X,red,n-1];
		irr2=DeleteCases[Flatten[Table[Prepend[#,w]&/@Cases[irr1,{Repeated[_,{n-1}]}],{w,X}],1],Alternatives@@Map[Append[#,___]&,red]];
		If[Head[level]===List,irr2,Join[irr1,irr2]]
	]
]


ReducibleWordQ[w_List,reduciblewords_List]:=
Module[{red,spec},
	spec=Map[Rule[#[[1]],Alternatives[#[[1]],Sequence@@#[[2]]]]&,Specialization];
	red=Map[{___,Sequence@@#,___}&,Replace[reduciblewords,spec,{2}]];
	MatchQ[w,Alternatives@@red]
]


(* ::Subsubsection:: *)
(*Generate ambiguities*)


(* ::Text:: *)
(*Overlap[v,w] returns a list of all {a,b,c} such that a=vb=cw with |b|<|w|, |c|<|v|, |b|>0, and |c|>0.*)


Overlap[v_List,w_List]:=
Module[{i,overlaps={}},
	For[i=1,i<Length[v]&&i<Length[w],i++,
		If[Take[v,-i]===Take[w,i],
			AppendTo[overlaps,Overlap[Join[v,Drop[w,i]],Drop[w,i],Drop[v,-i]]]
		]
	];
	overlaps
]


(* ::Text:: *)
(*Inclusion[v,w] returns a list of all {a,b,c} such that a=v=bwc with |b|+|c|>0.*)


Inclusion[v_List,w_List]:=
Module[{i,inclusions={}},
	If[Length[w]<Length[v],
		For[i=1,i+Length[w]-1<=Length[v],i++,
			If[v[[i;;(i+Length[w]-1)]]===w,
				AppendTo[inclusions,Inclusion[v,v[[1;;(i-1)]],v[[(i+Length[w]);;]]]]
			]
		]
	];
	inclusions
]


(* ::Text:: *)
(*SpecialOverlap[v,w] and SpecialInclusion[v,w] behave like Overlap[v,w] and Inclusion[v,w], but take specialization into account. Only those overlaps and inclusions are returned that cannot be obtained by Overlap[v,w] and Inclusion[v,w].*)


SpecialOverlap[v_List,w_List]:=
Module[{i,overlaps={},spec},
	For[i=1,i<Length[v]&&i<Length[w],i++,
		spec=Specialize[Take[v,-i],Take[w,i]];
		If[spec=!={{},{}},
			AppendTo[overlaps,SpecialOverlap[Join[Drop[v,-i],ReplacePart[Take[v,-i],spec[[1]]],Drop[w,i]],Drop[w,i],Drop[v,-i]]]
		]
	];
	overlaps
]


SpecialInclusion[v_List,w_List]:=
Module[{i,inclusions={},spec},
	If[Length[w]<Length[v],
		For[i=1,i+Length[w]-1<=Length[v],i++,
			spec=Specialize[v[[i;;(i+Length[w]-1)]],w];
			If[spec=!={{},{}},
				AppendTo[inclusions,SpecialInclusion[Join[v[[1;;(i-1)]],ReplacePart[w,spec[[2]]],v[[(i+Length[w]);;]]],v[[1;;(i-1)]],v[[(i+Length[w]);;]]]]
			]
		]
	];
	inclusions
]


(* ::Text:: *)
(*The following commands generate a list of all corresponding ambiguities.*)


GenerateOverlaps[l_List]:=Flatten[Outer[Overlap,l,l,1],2]


GenerateInclusions[l_List]:=Flatten[Outer[Inclusion,l,l,1],2]


GenerateSpecialOverlaps[l_List]:=Flatten[Outer[SpecialOverlap,l,l,1],2]


GenerateSpecialInclusions[l_List]:=Flatten[Outer[SpecialInclusion,l,l,1],2]


GenerateSpecialOverlaps[l1_List,l2_List]:=Flatten[Outer[{SpecialOverlap[#1,#2],SpecialOverlap[#2,#1]}&,l1,l2,1],3]


GenerateSpecialInclusions[l1_List,l2_List]:=Flatten[Outer[{SpecialInclusion[#1,#2],SpecialInclusion[#2,#1]}&,l1,l2,1],3]


GenerateAmbiguities[l_List]:=Join[GenerateOverlaps[l],GenerateInclusions[l],GenerateSpecialOverlaps[l],GenerateSpecialInclusions[l]]


(* ::Subsubsection:: *)
(*S-polynomials*)


SPoly[amb:(Alternatives[Overlap,Inclusion,SpecialOverlap,SpecialInclusion][_List,_List,_List]),sys:ReductionSystem]:=
Module[{h1,h2,i,n1,n2,t,x,pos,tmp},
	t=amb[[1]];
	n1=Length[amb[[2]]];n2=Length[amb[[3]]];
	Do[
		pos=Position[t,x,{1},Heads->False];
		tmp=Cases[CyclicModules,{x,_}];
		If[tmp==={},
			tmp=Table[Rule[pos[[i]],x[i]],{i,Length[pos]}]
		,
			tmp=Table[Rule[pos[[i]],tmp[[1,2]]],{i,Length[pos]}]
		];
		t=ReplacePart[t,tmp]
	,{x,Union[amb[[1]]]}];
	Switch[amb[[0]],
		Alternatives[Overlap,SpecialOverlap],
			h1=FindRule[Drop[amb[[1]],-n1],sys];
			If[h1==={},h1=$Failed,h1=h1[[1,2]]];
			h2=FindRule[Drop[amb[[1]],n2],sys];
			If[h2==={},h2=$Failed,h2=h2[[1,2]]];
			Prod[h1@@Drop[t,-n1],Sequence@@Take[t,-n1]]-Prod[Sequence@@Take[t,n2],h2@@Drop[t,n2]],
		Alternatives[Inclusion,SpecialInclusion],
			h1=FindRule[amb[[1]],sys];
			If[h1==={},h1=$Failed,h1=h1[[1,2]]];
			h2=FindRule[Take[amb[[1]],{n1+1,-n2-1}],sys];
			If[h2==={},h2=$Failed,h2=h2[[1,2]]];
			(h1@@t)-Prod[Sequence@@Take[t,n1],h2@@Take[t,{n1+1,-n2-1}],Sequence@@Take[t,-n2]]
	]
]


CheckResolvability[sys:ReductionSystem,OptionsPattern[{Print->False,Count->False}]]:=
Module[{amb,spol,i,p,c},
	p=(OptionValue[Print]===True);
	c=(OptionValue[Count]===True);
	amb=GenerateAmbiguities[ExtractReducibleWords[sys]];
	If[c,Print[Length[amb]," ambiguities in total"]];
	spol=Map[SPoly[#,sys]&,amb];
	If[c,Print[Length[Cases[spol,0]]," ambiguities have all S-polynomials equal to zero"]];
	If[p,
		spol=Table[Print[i,": ",amb[[i]]];Expand[ApplyRules[spol[[i]],sys,Print->True]],{i,Length[amb]}]
	,
		spol=Expand[ApplyRules[spol,sys]]
	];
	If[c,Print[Length[Cases[spol,0]]," ambiguities are resolvable"]];
	amb=Extract[amb,Position[spol,Except[0],{1},Heads->False]];
	Transpose[{amb,DeleteCases[spol,0]}]
]


(* ::Subsubsection:: *)
(*Output*)


Copyright[a_String,b___String]:=Print[StringJoin[Prepend[{"\n",#}&/@{b},a]]]


Copyright[
    "Package TenReS version 0.2.3",
    "written by Clemens G. Raab",
    "Copyright 2016, RICAM, Austrian Academy of Sciences"];


End[]


EndPackage[]


?TenReS
