(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 10.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       158,          7]
NotebookDataLength[     57180,       1715]
NotebookOptionsPosition[     51750,       1531]
NotebookOutlinePosition[     52493,       1559]
CellTagsIndexPosition[     52407,       1554]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell[BoxData[
 RowBox[{"<<", "\"\<TenReS.m\>\""}]], "Input",
 CellChangeTimes->{{3.669468187220231*^9, 3.669468197002075*^9}, {
   3.66946847007839*^9, 3.669468484206896*^9}, {3.669468518675226*^9, 
   3.6694685189880333`*^9}, {3.669543796712756*^9, 3.669543797277092*^9}, 
   3.6701495140786753`*^9}],

Cell[CellGroupData[{

Cell[BoxData["\<\"Package TenReS version 0.2.3\\nwritten by Clemens G. \
Raab\\nCopyright 2016, RICAM, Austrian Academy of Sciences\"\>"], "Print",
 CellChangeTimes->{3.670410670701865*^9, 3.6704107258157587`*^9, 
  3.670411548552721*^9}],

Cell[BoxData[
 StyleBox["\<\"A few definitions need to be made by the user. Type ?CoeffQ, \
?Specialization, and ?CyclicModules for more information.\"\>", 
  "MSG"]], "Print", "PrintUsage",
 CellChangeTimes->{3.670411548679454*^9},
 CellTags->"Info3670418748-3520926"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "CoeffQ"}]], "Input",
 CellChangeTimes->{{3.66954397515504*^9, 3.6695439875325527`*^9}, {
  3.669544086129202*^9, 3.6695440875090714`*^9}}],

Cell[BoxData[
 StyleBox["\<\"TenReS`CoeffQ[k_] should check whether k is in the ring over \
which the tensor product is formed.\"\>", "MSG"]], "Print", "PrintUsage",
 CellChangeTimes->{3.670411548784803*^9},
 CellTags->"Info3670418748-3520926"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "Specialization"}]], "Input",
 CellChangeTimes->{{3.6695440908186913`*^9, 3.669544098846047*^9}}],

Cell[BoxData[
 StyleBox["\<\"TenReS`Specialization should store a list of the form \
{Rule[y1,X[y1]],Rule[y2,X[y2]],...} specifying the refinement of the \
alphabet.\"\>", "MSG"]], "Print", "PrintUsage",
 CellChangeTimes->{3.67041154888561*^9},
 CellTags->"Info3670418748-3520926"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "CyclicModules"}]], "Input",
 CellChangeTimes->{{3.66954411423869*^9, 3.669544121706465*^9}}],

Cell[BoxData[
 StyleBox["\<\"TenReS`CyclicModules should store a list of pairs {x,gen} \
where the module corresponding to the letter x is generated by the element \
gen.\"\>", "MSG"]], "Print", "PrintUsage",
 CellChangeTimes->{3.670411548984758*^9},
 CellTags->"Info3670418748-3520926"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Definitions for differential operators", "Subsection",
 CellChangeTimes->{{3.669969699200149*^9, 3.669969710574629*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"CoeffQ", "=", 
   SubscriptBox["MemberQ", "K"]}], ";"}]], "Input",
 CellChangeTimes->{3.669455841818426*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{"Specialization", "=", 
   RowBox[{"{", 
    RowBox[{"F", "\[Rule]", 
     RowBox[{"{", 
      RowBox[{"K", ",", 
       OverscriptBox["F", "~"]}], "}"}]}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{
  3.6694556672615433`*^9, {3.669456324561751*^9, 3.6694563559371567`*^9}, {
   3.669457701410192*^9, 3.669457729353508*^9}, 3.6695435464591618`*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{"CyclicModules", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"K", ",", "1"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"D", ",", "Diff"}], "}"}]}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.66945752314161*^9, 3.6694575905696487`*^9}, {
  3.669522055347418*^9, 3.669522071897049*^9}, {3.66954139761335*^9, 
  3.6695413988525763`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   SubscriptBox["MemberQ", "K"], "[", "f_", "]"}], ":=", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{"Diff", "[", "f", "]"}], "===", "0"}], ")"}]}]], "Input",
 CellChangeTimes->{{3.669455734551148*^9, 3.669455760199451*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   SubscriptBox["MemberQ", "F"], "[", 
   RowBox[{"f_", "?", 
    SubscriptBox["MemberQ", "K"]}], "]"}], ":=", 
  "True"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   SubscriptBox["MemberQ", "F"], "[", 
   RowBox[{"F", "[", "_Integer", "]"}], "]"}], ":=", 
  "True"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   SubscriptBox["MemberQ", "F"], "[", 
   RowBox[{"Diff", "[", 
    RowBox[{"f_", "?", 
     SubscriptBox["MemberQ", "F"]}], "]"}], "]"}], ":=", 
  "True"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   SubscriptBox["MemberQ", "F"], "[", "f_mul", "]"}], ":=", 
  "True"}]}], "Input",
 CellChangeTimes->{{3.669455790328519*^9, 3.6694557970644712`*^9}, {
  3.669543619267727*^9, 3.669543632574329*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   SubscriptBox["MemberQ", "D"], "[", "Diff", "]"}], ":=", "True"}]], "Input",\

 CellChangeTimes->{{3.669455877198039*^9, 3.669455894498609*^9}, 
   3.669543189868985*^9}],

Cell[CellGroupData[{

Cell["Computation in the module", "Subsubsection",
 CellChangeTimes->{{3.6697321331548977`*^9, 3.669732144109561*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"mul", "[", 
   RowBox[{"a___", ",", "b_Plus", ",", "c___"}], "]"}], ":=", 
  RowBox[{"(", 
   RowBox[{
    RowBox[{
     RowBox[{"mul", "[", 
      RowBox[{"a", ",", "#", ",", "c"}], "]"}], "&"}], "/@", "b"}], 
   ")"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"mul", "[", 
   RowBox[{"a___", ",", "b_Integer", ",", "c__"}], "]"}], ":=", 
  RowBox[{"b", " ", 
   RowBox[{"mul", "[", 
    RowBox[{"a", ",", "c"}], "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"mul", "[", 
   RowBox[{"a__", ",", "b_Integer", ",", "c___"}], "]"}], ":=", 
  RowBox[{"b", " ", 
   RowBox[{"mul", "[", 
    RowBox[{"a", ",", "c"}], "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"mul", "[", 
   RowBox[{"a___", ",", 
    RowBox[{"d_Integer", "*", "b_"}], ",", "c___"}], "]"}], ":=", 
  RowBox[{"d", " ", 
   RowBox[{"mul", "[", 
    RowBox[{"a", ",", "b", ",", "c"}], "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"mul", "[", 
   RowBox[{"a___", ",", 
    RowBox[{"mul", "[", "b__", "]"}], ",", "c___"}], "]"}], ":=", 
  RowBox[{"mul", "[", 
   RowBox[{"a", ",", "b", ",", "c"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"mul", "[", "a_", "]"}], ":=", "a"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"mul", "[", 
   RowBox[{"a___", ",", 
    RowBox[{"c_", "?", 
     SubscriptBox["MemberQ", "K"]}], ",", "b___"}], "]"}], ":=", 
  RowBox[{"c", " ", 
   RowBox[{"mul", "[", 
    RowBox[{"a", ",", "b"}], "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"mul", "[", 
   RowBox[{"a___", ",", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"c_", "?", 
       SubscriptBox["MemberQ", "K"]}], ")"}], "f_"}], ",", "b___"}], "]"}], ":=", 
  RowBox[{"c", " ", 
   RowBox[{"mul", "[", 
    RowBox[{"a", ",", "f", ",", "b"}], "]"}]}]}]}], "Input",
 CellChangeTimes->{{3.66945583039043*^9, 3.6694558329265003`*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Diff", "[", "a_mul", "]"}], ":=", 
  RowBox[{"Sum", "[", 
   RowBox[{
    RowBox[{"MapAt", "[", 
     RowBox[{"Diff", ",", "a", ",", "i"}], "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"i", ",", 
      RowBox[{"Length", "[", "a", "]"}]}], "}"}]}], 
   "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Diff", "[", 
   RowBox[{"f_", "?", "NumericQ"}], "]"}], ":=", "0"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Diff", "[", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"f_", "?", 
      SubscriptBox["MemberQ", "K"]}], ")"}], "g_"}], "]"}], ":=", 
  RowBox[{"f", " ", 
   RowBox[{"Diff", "[", "g", "]"}]}]}]}], "Input",
 CellChangeTimes->{{3.669543586839528*^9, 3.6695435872970333`*^9}, {
  3.6701506211643353`*^9, 3.670150621443488*^9}}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Reduction system", "Subsection",
 CellChangeTimes->{{3.6697320805645657`*^9, 3.669732083293178*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   SubscriptBox["h", "FF"], "[", 
   RowBox[{"f_", ",", "g_"}], "]"}], ":=", 
  RowBox[{"Prod", "[", 
   RowBox[{"mul", "[", 
    RowBox[{"f", ",", "g"}], "]"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.669455480630855*^9, 3.669455481132794*^9}, {
  3.669456030364386*^9, 3.669456033724615*^9}, {3.669467853104024*^9, 
  3.669467859478427*^9}, {3.669467923694378*^9, 3.6694679513988333`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   SubscriptBox["h", "DF"], "[", 
   RowBox[{"Diff", ",", "f_"}], "]"}], ":=", 
  RowBox[{
   RowBox[{"Prod", "[", 
    RowBox[{"Diff", "[", "f", "]"}], "]"}], "+", 
   RowBox[{"Prod", "[", 
    RowBox[{"f", ",", "Diff"}], "]"}]}]}]], "Input",
 CellChangeTimes->{
  3.669456031556509*^9, {3.669467961987013*^9, 3.669467972655408*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    SubscriptBox["h", "K"], "[", "1", "]"}], ":=", 
   RowBox[{"Prod", "[", "]"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.669455469516754*^9, 3.669455472579659*^9}, 
   3.669467979465959*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{"RedSys", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"F", ",", "F"}], "}"}], ",", 
       SubscriptBox["h", "FF"]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"D", ",", "F"}], "}"}], ",", 
       SubscriptBox["h", "DF"]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", "K", "}"}], ",", 
       SubscriptBox["h", "K"]}], "}"}]}], "}"}]}], ";"}]], "Input",
 CellChangeTimes->{{3.669455239519739*^9, 3.669455242896411*^9}, {
   3.669455492727862*^9, 3.669455570280819*^9}, {3.66946799223446*^9, 
   3.66946800480234*^9}, 3.669542219531169*^9}],

Cell[CellGroupData[{

Cell["Applying the reduction system", "Subsubsection",
 CellChangeTimes->{{3.670411601272018*^9, 3.670411609801528*^9}}],

Cell[TextData[{
 "Reduce ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["f", "1"], "\[CircleTimes]", 
    RowBox[{"\[PartialD]", 
     RowBox[{"\[CircleTimes]", 
      SubscriptBox["f", "2"]}]}]}], TraditionalForm]],
  FormatType->"TraditionalForm"],
 " to its irreducible form ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       SubscriptBox["f", "1"], 
       SubscriptBox["f", "2"]}], ")"}], "\[CircleTimes]", 
     RowBox[{"\[PartialD]", 
      RowBox[{"+", 
       SubscriptBox["f", "1"]}]}]}], 
    RowBox[{"\[PartialD]", 
     SubscriptBox["f", "2"]}]}], TraditionalForm]],
  FormatType->"TraditionalForm"]
}], "Text",
 CellChangeTimes->{{3.670411857681672*^9, 3.670411961325132*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ApplyRules", "[", 
  RowBox[{
   RowBox[{"Prod", "[", 
    RowBox[{
     RowBox[{"F", "[", "1", "]"}], ",", "Diff", ",", 
     RowBox[{"F", "[", "2", "]"}]}], "]"}], ",", "RedSys"}], "]"}]], "Input",
 CellChangeTimes->{{3.669714574442025*^9, 3.669714611160872*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"Prod", "[", 
   RowBox[{"mul", "[", 
    RowBox[{
     RowBox[{"F", "[", "1", "]"}], ",", 
     RowBox[{"Diff", "[", 
      RowBox[{"F", "[", "2", "]"}], "]"}]}], "]"}], "]"}], "+", 
  RowBox[{"Prod", "[", 
   RowBox[{
    RowBox[{"mul", "[", 
     RowBox[{
      RowBox[{"F", "[", "1", "]"}], ",", 
      RowBox[{"F", "[", "2", "]"}]}], "]"}], ",", "Diff"}], "]"}]}]], "Output",\

 CellChangeTimes->{
  3.669714612042076*^9, 3.6697253989864197`*^9, 3.669725708730283*^9, 
   3.6697258940137367`*^9, 3.669726860360867*^9, 3.6697270521573563`*^9, 
   3.669727561405106*^9, 3.669727617668042*^9, 3.669727837918211*^9, {
   3.669727873302668*^9, 3.669727896853879*^9}, 3.6697279348032837`*^9, 
   3.669728070323593*^9, 3.669728313570209*^9, 3.669728635602598*^9, 
   3.669729044981337*^9, 3.669970316642314*^9, 3.669970417894518*^9, 
   3.670073453953003*^9, 3.670073645820282*^9, 3.6701505479153433`*^9, 
   3.6701507053604593`*^9, 3.6704106233657846`*^9, 3.670410671362052*^9, 
   3.670410726532464*^9, 3.6704115492048492`*^9}]
}, Open  ]],

Cell[TextData[{
 "Show the length of the longest chain of reductions on any particular \
subexpression, for example:\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["f", "1"], "\[CircleTimes]", 
     RowBox[{"\[PartialD]", 
      RowBox[{"\[CircleTimes]", 
       SubscriptBox["f", "2"]}]}]}], "\[Rule]", 
    RowBox[{
     SubscriptBox["f", "1"], "\[CircleTimes]", 
     RowBox[{"(", 
      RowBox[{
       SubscriptBox["f", "2"], "\[CircleTimes]", 
       RowBox[{"\[PartialD]", 
        RowBox[{"+", 
         RowBox[{"\[PartialD]", 
          SubscriptBox["f", "2"]}]}]}]}], ")"}]}]}], TraditionalForm]]],
 "\n",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["f", "1"], "\[CircleTimes]", 
     SubscriptBox["f", "2"], "\[CircleTimes]", "\[PartialD]"}], "\[Rule]", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       SubscriptBox["f", "1"], 
       SubscriptBox["f", "2"]}], ")"}], "\[CircleTimes]", "\[PartialD]"}]}], 
   TraditionalForm]],
  FormatType->"TraditionalForm"]
}], "Text",
 CellChangeTimes->{{3.67041163784116*^9, 3.6704116775295477`*^9}, {
  3.670411787840767*^9, 3.670411795934536*^9}, {3.670412023186578*^9, 
  3.6704121633316402`*^9}, {3.670412205651499*^9, 3.670412242246172*^9}, {
  3.6704123264337893`*^9, 3.67041232650604*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ApplyRules", "[", 
  RowBox[{
   RowBox[{"Prod", "[", 
    RowBox[{
     RowBox[{"F", "[", "1", "]"}], ",", "Diff", ",", 
     RowBox[{"F", "[", "2", "]"}]}], "]"}], ",", "RedSys", ",", 
   RowBox[{"Count", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.6704115233597593`*^9, 3.6704115240992947`*^9}}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"2", "\[InvisibleSpace]", "\<\" steps\"\>"}],
  SequenceForm[2, " steps"],
  Editable->False]], "Print",
 CellChangeTimes->{3.670411549250718*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{"Prod", "[", 
   RowBox[{"mul", "[", 
    RowBox[{
     RowBox[{"F", "[", "1", "]"}], ",", 
     RowBox[{"Diff", "[", 
      RowBox[{"F", "[", "2", "]"}], "]"}]}], "]"}], "]"}], "+", 
  RowBox[{"Prod", "[", 
   RowBox[{
    RowBox[{"mul", "[", 
     RowBox[{
      RowBox[{"F", "[", "1", "]"}], ",", 
      RowBox[{"F", "[", "2", "]"}]}], "]"}], ",", "Diff"}], "]"}]}]], "Output",\

 CellChangeTimes->{3.6704115492578287`*^9}]
}, Open  ]],

Cell["Print reduction rules as they are applied:", "Text",
 CellChangeTimes->{{3.670411682361391*^9, 3.670411700265997*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ApplyRules", "[", 
  RowBox[{
   RowBox[{"Prod", "[", 
    RowBox[{
     RowBox[{"F", "[", "1", "]"}], ",", "Diff", ",", 
     RowBox[{"F", "[", "2", "]"}]}], "]"}], ",", "RedSys", ",", 
   RowBox[{"Print", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.670410577938178*^9, 3.670410579986349*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"D", ",", "F"}], "}"}], ",", 
   SubscriptBox["h", "DF"]}], "}"}]], "Print",
 CellChangeTimes->{3.670410671399857*^9, 3.670410726587152*^9, 
  3.670411549301103*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"F", ",", "F"}], "}"}], ",", 
   SubscriptBox["h", "FF"]}], "}"}]], "Print",
 CellChangeTimes->{3.670410671399857*^9, 3.670410726587152*^9, 
  3.670411549311762*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"F", ",", "F"}], "}"}], ",", 
   SubscriptBox["h", "FF"]}], "}"}]], "Print",
 CellChangeTimes->{3.670410671399857*^9, 3.670410726587152*^9, 
  3.670411549320209*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"Prod", "[", 
   RowBox[{"mul", "[", 
    RowBox[{
     RowBox[{"F", "[", "1", "]"}], ",", 
     RowBox[{"Diff", "[", 
      RowBox[{"F", "[", "2", "]"}], "]"}]}], "]"}], "]"}], "+", 
  RowBox[{"Prod", "[", 
   RowBox[{
    RowBox[{"mul", "[", 
     RowBox[{
      RowBox[{"F", "[", "1", "]"}], ",", 
      RowBox[{"F", "[", "2", "]"}]}], "]"}], ",", "Diff"}], "]"}]}]], "Output",\

 CellChangeTimes->{3.670410623406344*^9, 3.6704106714235697`*^9, 
  3.6704107266139727`*^9, 3.670411549328405*^9}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Ambiguities and confluence", "Subsection",
 CellChangeTimes->{{3.669732094245129*^9, 3.6697321007895727`*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ambiguities", "=", 
  RowBox[{"GenerateAmbiguities", "[", 
   RowBox[{"ExtractReducibleWords", "[", "RedSys", "]"}], "]"}]}]], "Input",
 CellChangeTimes->{{3.669456191617311*^9, 3.669456228396771*^9}, {
  3.669541582677245*^9, 3.66954159305774*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"Overlap", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"F", ",", "F", ",", "F"}], "}"}], ",", 
     RowBox[{"{", "F", "}"}], ",", 
     RowBox[{"{", "F", "}"}]}], "]"}], ",", 
   RowBox[{"Overlap", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"D", ",", "F", ",", "F"}], "}"}], ",", 
     RowBox[{"{", "F", "}"}], ",", 
     RowBox[{"{", "D", "}"}]}], "]"}], ",", 
   RowBox[{"SpecialInclusion", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"K", ",", "F"}], "}"}], ",", 
     RowBox[{"{", "}"}], ",", 
     RowBox[{"{", "F", "}"}]}], "]"}], ",", 
   RowBox[{"SpecialInclusion", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"F", ",", "K"}], "}"}], ",", 
     RowBox[{"{", "F", "}"}], ",", 
     RowBox[{"{", "}"}]}], "]"}], ",", 
   RowBox[{"SpecialInclusion", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"D", ",", "K"}], "}"}], ",", 
     RowBox[{"{", "D", "}"}], ",", 
     RowBox[{"{", "}"}]}], "]"}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.669456231096222*^9, 3.669541594449711*^9, 3.6695419463452463`*^9, 
   3.669542936852714*^9, 3.6695430245059853`*^9, 3.669543137457488*^9, 
   3.669543460659218*^9, 3.669543657524158*^9, 3.669544376072371*^9, 
   3.6695447436636057`*^9, 3.669725398760571*^9, 3.669725708521978*^9, 
   3.669725893811309*^9, 3.669726860037943*^9, 3.669727051939967*^9, 
   3.669727561179491*^9, 3.669727617447798*^9, 3.6697278377283278`*^9, {
   3.669727873080318*^9, 3.669727896630465*^9}, 3.669727934616555*^9, 
   3.669728070103115*^9, 3.669728313336794*^9, 3.669728635410191*^9, 
   3.669729044748362*^9, 3.6699703166734467`*^9, 3.6699704179442797`*^9, 
   3.670073454006796*^9, 3.670073645866108*^9, 3.670150547969976*^9, 
   3.670150705397942*^9, 3.6704106234564342`*^9, 3.670410671466243*^9, 
   3.6704107266565037`*^9, 3.670411549368331*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"spolys", "=", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"SPoly", "[", 
      RowBox[{"#", ",", "RedSys"}], "]"}], "&"}], ",", "ambiguities"}], 
   "]"}]}]], "Input",
 CellChangeTimes->{{3.669543046908551*^9, 3.6695430916312733`*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     RowBox[{"Prod", "[", 
      RowBox[{
       RowBox[{"F", "[", "1", "]"}], ",", 
       RowBox[{"mul", "[", 
        RowBox[{
         RowBox[{"F", "[", "2", "]"}], ",", 
         RowBox[{"F", "[", "3", "]"}]}], "]"}]}], "]"}]}], "+", 
    RowBox[{"Prod", "[", 
     RowBox[{
      RowBox[{"mul", "[", 
       RowBox[{
        RowBox[{"F", "[", "1", "]"}], ",", 
        RowBox[{"F", "[", "2", "]"}]}], "]"}], ",", 
      RowBox[{"F", "[", "3", "]"}]}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"-", 
     RowBox[{"Prod", "[", 
      RowBox[{"Diff", ",", 
       RowBox[{"mul", "[", 
        RowBox[{
         RowBox[{"F", "[", "1", "]"}], ",", 
         RowBox[{"F", "[", "2", "]"}]}], "]"}]}], "]"}]}], "+", 
    RowBox[{"Prod", "[", 
     RowBox[{
      RowBox[{"Diff", "[", 
       RowBox[{"F", "[", "1", "]"}], "]"}], ",", 
      RowBox[{"F", "[", "2", "]"}]}], "]"}], "+", 
    RowBox[{"Prod", "[", 
     RowBox[{
      RowBox[{"F", "[", "1", "]"}], ",", "Diff", ",", 
      RowBox[{"F", "[", "2", "]"}]}], "]"}]}], ",", "0", ",", "0", ",", 
   RowBox[{
    RowBox[{"-", 
     RowBox[{"Prod", "[", "Diff", "]"}]}], "+", 
    RowBox[{"Prod", "[", 
     RowBox[{"1", ",", "Diff"}], "]"}]}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.669543057883449*^9, 3.669543137502407*^9, 3.6695432073101597`*^9, 
   3.6695434607093763`*^9, 3.669543657574725*^9, 3.6695443761152077`*^9, 
   3.6695447437012177`*^9, 3.669725398784731*^9, 3.669725708558902*^9, 
   3.6697258938642073`*^9, 3.6697268600762577`*^9, 3.669727051992878*^9, 
   3.66972756122121*^9, 3.669727617500527*^9, 3.669727837757216*^9, {
   3.669727873122754*^9, 3.6697278966724777`*^9}, 3.669727934645699*^9, 
   3.669728070137624*^9, 3.669728313387897*^9, 3.669728635450589*^9, 
   3.669729044802741*^9, 3.6699703167255697`*^9, 3.669970417999937*^9, 
   3.6700734540527153`*^9, 3.6700736459150553`*^9, 3.670150548019992*^9, 
   3.670150705448503*^9, 3.670410623504387*^9, 3.67041067151608*^9, 
   3.670410726706169*^9, 3.67041154941759*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ApplyRules", "[", 
  RowBox[{"spolys", ",", "RedSys"}], "]"}]], "Input",
 CellChangeTimes->{{3.669543104928244*^9, 3.66954311177555*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}]], "Output",
 CellChangeTimes->{
  3.669543137558256*^9, 3.669543208762939*^9, 3.669543460758832*^9, 
   3.669543657624312*^9, 3.6695443761610947`*^9, 3.66954474375099*^9, 
   3.6697253988292217`*^9, 3.669725708614339*^9, 3.669725893917316*^9, 
   3.6697268601161547`*^9, 3.669727052041903*^9, 3.669727561271035*^9, 
   3.669727617544643*^9, 3.669727837802888*^9, {3.6697278731730347`*^9, 
   3.669727896723413*^9}, 3.669727934685145*^9, 3.669728070190763*^9, 
   3.669728313437368*^9, 3.669728635505863*^9, 3.6697290448532467`*^9, 
   3.669970316780778*^9, 3.669970418049556*^9, 3.670073454109086*^9, 
   3.670073645965499*^9, 3.670150548070013*^9, 3.67015070549857*^9, 
   3.6704106235406446`*^9, 3.670410671567175*^9, 3.670410726756312*^9, 
   3.6704115494670763`*^9}]
}, Open  ]],

Cell["List those ambiguiteis that could not be resolved:", "Text",
 CellChangeTimes->{{3.670412522745953*^9, 3.6704125399065332`*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CheckResolvability", "[", "RedSys", "]"}]], "Input",
 CellChangeTimes->{{3.669523440842567*^9, 3.6695234464030657`*^9}, {
  3.669523506381168*^9, 3.669523508485373*^9}, {3.6695428513509703`*^9, 
  3.6695428559406767`*^9}}],

Cell[BoxData[
 RowBox[{"{", "}"}]], "Output",
 CellChangeTimes->{
  3.669542944145224*^9, 3.6695429752402897`*^9, 3.669543024609541*^9, 
   3.669543137617339*^9, 3.6695432100984373`*^9, 3.669543460819775*^9, 
   3.66954365768437*^9, 3.6695443762255707`*^9, 3.669544743816216*^9, 
   3.669725398905312*^9, 3.669725708690431*^9, 3.669725893967554*^9, 
   3.669726860190287*^9, 3.669727052118972*^9, 3.66972756135224*^9, 
   3.6697276176270943`*^9, 3.66972783788251*^9, {3.669727873249729*^9, 
   3.669727896802627*^9}, 3.669727934763357*^9, 3.669728070273075*^9, 
   3.66972831351866*^9, 3.6697286355615*^9, 3.669729044932165*^9, 
   3.669970316858383*^9, 3.669970418132756*^9, 3.670073454178801*^9, 
   3.670073646044902*^9, 3.6701505481474953`*^9, 3.670150705573895*^9, 
   3.670410623575989*^9, 3.670410671616116*^9, 3.6704107268065443`*^9, 
   3.670411549517928*^9}]
}, Open  ]],

Cell["Give some statistics on the number of ambiguities:", "Text",
 CellChangeTimes->{{3.670412393726863*^9, 3.670412405519305*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CheckResolvability", "[", 
  RowBox[{"RedSys", ",", 
   RowBox[{"Count", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.6704105899943113`*^9, 3.670410592826702*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"5", "\[InvisibleSpace]", "\<\" ambiguities in total\"\>"}],
  SequenceForm[5, " ambiguities in total"],
  Editable->False]], "Print",
 CellChangeTimes->{3.670410671666986*^9, 3.670410726856563*^9, 
  3.670411549626824*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{
  "2", "\[InvisibleSpace]", "\<\" ambiguities have all S-polynomials equal to \
zero\"\>"}],
  SequenceForm[2, " ambiguities have all S-polynomials equal to zero"],
  Editable->False]], "Print",
 CellChangeTimes->{3.670410671666986*^9, 3.670410726856563*^9, 
  3.670411549637477*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"5", "\[InvisibleSpace]", "\<\" ambiguities are resolvable\"\>"}],
  SequenceForm[5, " ambiguities are resolvable"],
  Editable->False]], "Print",
 CellChangeTimes->{3.670410671666986*^9, 3.670410726856563*^9, 
  3.670411549649797*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", "}"}]], "Output",
 CellChangeTimes->{3.670410623624846*^9, 3.670410671694219*^9, 
  3.6704107268818903`*^9, 3.6704115496603813`*^9}]
}, Open  ]],

Cell["Show for each ambiguity how its S-polynomials are reduced:", "Text",
 CellChangeTimes->{{3.670412413230474*^9, 3.670412455831687*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CheckResolvability", "[", 
  RowBox[{"RedSys", ",", 
   RowBox[{"Print", "\[Rule]", "True"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.670410596706465*^9, 3.670410599282734*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"1", "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
   RowBox[{"Overlap", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"F", ",", "F", ",", "F"}], "}"}], ",", 
     RowBox[{"{", "F", "}"}], ",", 
     RowBox[{"{", "F", "}"}]}], "]"}]}],
  SequenceForm[1, ": ", 
   TenReS`Overlap[{$CellContext`F, $CellContext`F, $CellContext`F}, \
{$CellContext`F}, {$CellContext`F}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.6704106718120117`*^9, 3.67041072692109*^9, 
  3.6704115497016983`*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"F", ",", "F"}], "}"}], ",", 
   SubscriptBox["h", "FF"]}], "}"}]], "Print",
 CellChangeTimes->{3.6704106718120117`*^9, 3.67041072692109*^9, 
  3.670411549711018*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"F", ",", "F"}], "}"}], ",", 
   SubscriptBox["h", "FF"]}], "}"}]], "Print",
 CellChangeTimes->{3.6704106718120117`*^9, 3.67041072692109*^9, 
  3.670411549719393*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"2", "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
   RowBox[{"Overlap", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"D", ",", "F", ",", "F"}], "}"}], ",", 
     RowBox[{"{", "F", "}"}], ",", 
     RowBox[{"{", "D", "}"}]}], "]"}]}],
  SequenceForm[2, ": ", 
   TenReS`Overlap[{D, $CellContext`F, $CellContext`F}, {$CellContext`F}, {D}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.6704106718120117`*^9, 3.67041072692109*^9, 
  3.670411549728025*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"D", ",", "F"}], "}"}], ",", 
   SubscriptBox["h", "DF"]}], "}"}]], "Print",
 CellChangeTimes->{3.6704106718120117`*^9, 3.67041072692109*^9, 
  3.6704115497364807`*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"F", ",", "F"}], "}"}], ",", 
   SubscriptBox["h", "FF"]}], "}"}]], "Print",
 CellChangeTimes->{3.6704106718120117`*^9, 3.67041072692109*^9, 
  3.670411549744849*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"D", ",", "F"}], "}"}], ",", 
   SubscriptBox["h", "DF"]}], "}"}]], "Print",
 CellChangeTimes->{3.6704106718120117`*^9, 3.67041072692109*^9, 
  3.670411549753777*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"F", ",", "F"}], "}"}], ",", 
   SubscriptBox["h", "FF"]}], "}"}]], "Print",
 CellChangeTimes->{3.6704106718120117`*^9, 3.67041072692109*^9, 
  3.670411549762417*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"F", ",", "F"}], "}"}], ",", 
   SubscriptBox["h", "FF"]}], "}"}]], "Print",
 CellChangeTimes->{3.6704106718120117`*^9, 3.67041072692109*^9, 
  3.6704115497708273`*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"3", "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
   RowBox[{"SpecialInclusion", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"K", ",", "F"}], "}"}], ",", 
     RowBox[{"{", "}"}], ",", 
     RowBox[{"{", "F", "}"}]}], "]"}]}],
  SequenceForm[3, ": ", 
   TenReS`SpecialInclusion[{K, $CellContext`F}, {}, {$CellContext`F}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.6704106718120117`*^9, 3.67041072692109*^9, 
  3.670411549779099*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"4", "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
   RowBox[{"SpecialInclusion", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"F", ",", "K"}], "}"}], ",", 
     RowBox[{"{", "F", "}"}], ",", 
     RowBox[{"{", "}"}]}], "]"}]}],
  SequenceForm[4, ": ", 
   TenReS`SpecialInclusion[{$CellContext`F, K}, {$CellContext`F}, {}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.6704106718120117`*^9, 3.67041072692109*^9, 
  3.670411549787368*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"5", "\[InvisibleSpace]", "\<\": \"\>", "\[InvisibleSpace]", 
   RowBox[{"SpecialInclusion", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"D", ",", "K"}], "}"}], ",", 
     RowBox[{"{", "D", "}"}], ",", 
     RowBox[{"{", "}"}]}], "]"}]}],
  SequenceForm[5, ": ", 
   TenReS`SpecialInclusion[{D, K}, {D}, {}]],
  Editable->False]], "Print",
 CellChangeTimes->{3.6704106718120117`*^9, 3.67041072692109*^9, 
  3.670411549795619*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", "K", "}"}], ",", 
   SubscriptBox["h", "K"]}], "}"}]], "Print",
 CellChangeTimes->{3.6704106718120117`*^9, 3.67041072692109*^9, 
  3.670411549803866*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{"{", "}"}]], "Output",
 CellChangeTimes->{3.6704106236758213`*^9, 3.670410671861782*^9, 
  3.6704107270407457`*^9, 3.670411549811983*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Irreducible words", "Subsection",
 CellChangeTimes->{{3.669732052670609*^9, 3.6697320588059*^9}}],

Cell["\<\
Irreducible words up to a certain length can be generated by specifying the \
alphabet and the list of reducible words.\
\>", "Text",
 CellChangeTimes->{{3.669732172951088*^9, 3.669732191055724*^9}, {
  3.669732233170917*^9, 3.669732289990765*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IrreducibleWords", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"K", ",", 
     OverscriptBox["F", "~"], ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", "K", "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
       OverscriptBox["F", "~"], ",", 
       OverscriptBox["F", "~"]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"D", ",", 
       OverscriptBox["F", "~"]}], "}"}]}], "}"}], ",", "4"}], "]"}]], "Input",\

 CellChangeTimes->{{3.669725954929641*^9, 3.669726033562214*^9}, {
  3.669726108485805*^9, 3.669726122900999*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    OverscriptBox["F", "~"], "}"}], ",", 
   RowBox[{"{", "D", "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     OverscriptBox["F", "~"], ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"D", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     OverscriptBox["F", "~"], ",", "D", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"D", ",", "D", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     OverscriptBox["F", "~"], ",", "D", ",", "D", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"D", ",", "D", ",", "D", ",", "D"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.669725966006124*^9, 3.669726124574018*^9, 3.669726860385727*^9, 
   3.669727052209388*^9, 3.6697275614587193`*^9, 3.6697276177170267`*^9, 
   3.66972783796891*^9, {3.669727873330288*^9, 3.669727896906962*^9}, 
   3.669727934847313*^9, 3.669728070372097*^9, 3.6697283136215973`*^9, 
   3.669728635655736*^9, 3.669729045005549*^9, 3.6699703169137287`*^9, 
   3.66997041817747*^9, 3.670073454222694*^9, 3.6700736460797443`*^9, 
   3.6701505481872377`*^9, 3.670150705616951*^9, 3.670410623725403*^9, 
   3.670410671915111*^9, 3.670410727170442*^9, 3.670411549848198*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IrreducibleWords", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"F", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", "K", "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"F", ",", "F"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"D", ",", "F"}], "}"}]}], "}"}], ",", "4"}], "]"}]], "Input",
 CellChangeTimes->{{3.6697287980387163`*^9, 3.669728813990403*^9}, 
   3.669728994450241*^9, 3.669729052317298*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    OverscriptBox["F", "~"], "}"}], ",", 
   RowBox[{"{", "D", "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     OverscriptBox["F", "~"], ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"D", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     OverscriptBox["F", "~"], ",", "D", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"D", ",", "D", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     OverscriptBox["F", "~"], ",", "D", ",", "D", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"D", ",", "D", ",", "D", ",", "D"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.669728821952018*^9, 3.669728995003909*^9, {3.66972904505658*^9, 
   3.669729052989847*^9}, 3.6699703169614487`*^9, 3.6699704182333517`*^9, 
   3.6700734542557287`*^9, 3.670073646130591*^9, 3.6701505482388887`*^9, 
   3.670150705668269*^9, 3.670410623777905*^9, 3.670410671938418*^9, 
   3.6704107272197723`*^9, 3.670411549898492*^9}]
}, Open  ]],

Cell["\<\
They can also be generated by just providing the list of irreducible words, \
the alphabet is then automatically inferred.\
\>", "Text",
 CellChangeTimes->{{3.6697323317026577`*^9, 3.669732377717943*^9}, {
  3.6699695753831*^9, 3.669969579186245*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IrreducibleWords", "[", 
  RowBox[{
   RowBox[{"ExtractReducibleWords", "[", "RedSys", "]"}], ",", "4"}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.669726932655896*^9, 3.669726937550747*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    OverscriptBox["F", "~"], "}"}], ",", 
   RowBox[{"{", "D", "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     OverscriptBox["F", "~"], ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"D", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     OverscriptBox["F", "~"], ",", "D", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"D", ",", "D", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     OverscriptBox["F", "~"], ",", "D", ",", "D", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"D", ",", "D", ",", "D", ",", "D"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.66972693972723*^9, 3.669727052258359*^9, 3.669727561489119*^9, 
   3.6697276177668867`*^9, 3.669727838013205*^9, {3.66972787337176*^9, 
   3.669727896955318*^9}, 3.669727934885379*^9, 3.669728070420781*^9, 
   3.669728313673147*^9, 3.66972863570461*^9, 3.66972904525816*^9, 
   3.669970317016623*^9, 3.669970418283114*^9, 3.6700734543469477`*^9, 
   3.6700736461807613`*^9, 3.670150548288125*^9, 3.670150705716571*^9, 
   3.6704106238255243`*^9, 3.670410671982745*^9, 3.670410727270709*^9, 
   3.670411549948957*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Select", "[", 
  RowBox[{
   RowBox[{"Flatten", "[", 
    RowBox[{
     RowBox[{"Table", "[", 
      RowBox[{
       RowBox[{"Flatten", "[", 
        RowBox[{
         RowBox[{"Outer", "[", 
          RowBox[{"List", ",", 
           RowBox[{"Sequence", "@@", 
            RowBox[{"Table", "[", 
             RowBox[{
              RowBox[{"{", 
               RowBox[{"K", ",", 
                OverscriptBox["F", "~"], ",", "D"}], "}"}], ",", 
              RowBox[{"{", "i", "}"}]}], "]"}]}]}], "]"}], ",", 
         RowBox[{"i", "-", "1"}]}], "]"}], ",", 
       RowBox[{"{", 
        RowBox[{"i", ",", "4"}], "}"}]}], "]"}], ",", "1"}], "]"}], ",", 
   RowBox[{
    RowBox[{"!", 
     RowBox[{"ReducibleWordQ", "[", 
      RowBox[{"#", ",", 
       RowBox[{"ExtractReducibleWords", "[", "RedSys", "]"}]}], "]"}]}], 
    "&"}]}], "]"}]], "Input",
 CellChangeTimes->{{3.669730205553437*^9, 3.669730231901545*^9}, {
  3.66973047990239*^9, 3.669730491265916*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    OverscriptBox["F", "~"], "}"}], ",", 
   RowBox[{"{", "D", "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     OverscriptBox["F", "~"], ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"D", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     OverscriptBox["F", "~"], ",", "D", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"D", ",", "D", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     OverscriptBox["F", "~"], ",", "D", ",", "D", ",", "D"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"D", ",", "D", ",", "D", ",", "D"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.669730481065083*^9, 3.669730580311849*^9, 
  3.669970317149542*^9, 3.66997041841131*^9, 3.670073454385028*^9, 
  3.670073646231292*^9, 3.6701505484048347`*^9, 3.670150705767508*^9, 
  3.670410623925681*^9, 3.670410672032837*^9, 3.670410727321142*^9, 
  3.670411549998973*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Efficiency", "Subsubsection",
 CellChangeTimes->{{3.669969761962386*^9, 3.66996976405128*^9}}],

Cell["\<\
IrreducibleWords is more efficient than generating all words and selecting \
the irreducible ones:\
\>", "Text",
 CellChangeTimes->{{3.669730729177565*^9, 3.6697307413931017`*^9}, {
  3.669730776576186*^9, 3.669730787912549*^9}, {3.669731434185001*^9, 
  3.669731448552466*^9}, {3.669969607622334*^9, 3.669969610070149*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"IrreducibleWords", "[", 
   RowBox[{
    RowBox[{"ExtractReducibleWords", "[", "RedSys", "]"}], ",", 
    RowBox[{"{", "10", "}"}]}], "]"}], "//", "Timing"}]], "Input",
 CellChangeTimes->{{3.669730694755178*^9, 3.6697307006446733`*^9}, {
  3.669732205631575*^9, 3.6697322075672903`*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "0.00124099999999999991567856127971936075`3.1143716947783537", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       OverscriptBox["F", "~"], ",", "D", ",", "D", ",", "D", ",", "D", ",", 
       "D", ",", "D", ",", "D", ",", "D", ",", "D"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
      "D", ",", "D", ",", "D", ",", "D", ",", "D", ",", "D", ",", "D", ",", 
       "D", ",", "D", ",", "D"}], "}"}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.6697307013093767`*^9, 3.66973220962528*^9, 
  3.6699703171984367`*^9, 3.669970418466954*^9, 3.670073454418219*^9, 
  3.670073646280875*^9, 3.670150548456399*^9, 3.670150705876152*^9, 
  3.670410623973177*^9, 3.670410672083316*^9, 3.670410727371196*^9, 
  3.670411550049016*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Timing", "[", 
  RowBox[{"wordsoflength10", "=", 
   RowBox[{"Flatten", "[", 
    RowBox[{
     RowBox[{"Outer", "[", 
      RowBox[{"List", ",", 
       RowBox[{"Sequence", "@@", 
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"K", ",", 
            OverscriptBox["F", "~"], ",", "D"}], "}"}], ",", 
          RowBox[{"{", "10", "}"}]}], "]"}]}]}], "]"}], ",", "9"}], "]"}]}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.669730838534627*^9, 3.669730845908334*^9}, {
  3.669731882079315*^9, 3.6697319342252817`*^9}}],

Cell[BoxData[
 InterpretationBox[
  TagBox[
   FrameBox[GridBox[{
      {
       ItemBox[
        TagBox[
         RowBox[{"{", 
          RowBox[{
          "0.03975100000000000161071156412617710885`4.61994797178203", ",", 
           RowBox[{"{", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
              "K", ",", "K", ",", "K", ",", "K", ",", "K", ",", "K", ",", "K",
                ",", "K", ",", "K", ",", "K"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
              "K", ",", "K", ",", "K", ",", "K", ",", "K", ",", "K", ",", "K",
                ",", "K", ",", "K", ",", 
               OverscriptBox["F", "~"]}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
              "K", ",", "K", ",", "K", ",", "K", ",", "K", ",", "K", ",", "K",
                ",", "K", ",", "K", ",", "D"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
              "K", ",", "K", ",", "K", ",", "K", ",", "K", ",", "K", ",", "K",
                ",", "K", ",", 
               OverscriptBox["F", "~"], ",", "K"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
              "K", ",", "K", ",", "K", ",", "K", ",", "K", ",", "K", ",", "K",
                ",", "K", ",", 
               OverscriptBox["F", "~"], ",", 
               OverscriptBox["F", "~"]}], "}"}], ",", 
             TemplateBox[{"59039"},
              "OutputSizeLimit`Skeleton",
              DisplayFunction->(FrameBox[
                
                RowBox[{
                 "\" \[CenterEllipsis]\"", #, "\"\[CenterEllipsis] \""}], 
                Background -> GrayLevel[0.75], 
                BaseStyle -> {
                 "Deploy", FontColor -> GrayLevel[1], FontSize -> Smaller, 
                  ShowStringCharacters -> False}, BaselinePosition -> 
                Baseline, ContentPadding -> False, FrameMargins -> 1, 
                FrameStyle -> GrayLevel[0.75], RoundingRadius -> 7]& )], ",", 
             
             RowBox[{"{", 
              RowBox[{
              "D", ",", "D", ",", "D", ",", "D", ",", "D", ",", "D", ",", "D",
                ",", "D", ",", 
               OverscriptBox["F", "~"], ",", 
               OverscriptBox["F", "~"]}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
              "D", ",", "D", ",", "D", ",", "D", ",", "D", ",", "D", ",", "D",
                ",", "D", ",", 
               OverscriptBox["F", "~"], ",", "D"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
              "D", ",", "D", ",", "D", ",", "D", ",", "D", ",", "D", ",", "D",
                ",", "D", ",", "D", ",", "K"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
              "D", ",", "D", ",", "D", ",", "D", ",", "D", ",", "D", ",", "D",
                ",", "D", ",", "D", ",", 
               OverscriptBox["F", "~"]}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
              "D", ",", "D", ",", "D", ",", "D", ",", "D", ",", "D", ",", "D",
                ",", "D", ",", "D", ",", "D"}], "}"}]}], "}"}]}], "}"}],
         Short[#, 5]& ],
        BaseStyle->{Deployed -> False},
        StripOnInput->False]},
      {GridBox[{
         {
          TagBox[
           TooltipBox[
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource[
                "FEStrings", "sizeBriefExplanation"], StandardForm],
               ImageSizeCache->{58., {2., 8.}}],
              StripOnInput->False,
              DynamicUpdating->True], "SuggestionsBarCategory",
             StripOnInput->False],
            StyleBox[
             DynamicBox[
              ToBoxes[
               FEPrivate`FrontEndResource["FEStrings", "sizeExplanation"], 
               StandardForm]], DynamicUpdating -> True, StripOnInput -> 
             False]],
           Annotation[#, 
            Style[
             Dynamic[
              FEPrivate`FrontEndResource["FEStrings", "sizeExplanation"]], 
             DynamicUpdating -> True], "Tooltip"]& ], 
          ButtonBox[
           PaneSelectorBox[{False->
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource["FEStrings", "sizeShowLess"], 
                StandardForm],
               ImageSizeCache->{51., {0., 8.}}],
              StripOnInput->False,
              DynamicUpdating->True], "SuggestionsBarControl",
             StripOnInput->False], True->
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource["FEStrings", "sizeShowLess"], 
                StandardForm]],
              StripOnInput->False,
              DynamicUpdating->True], "SuggestionsBarControlActive",
             StripOnInput->False]}, Dynamic[
             CurrentValue["MouseOver"]],
            Alignment->Center,
            FrameMargins->0,
            ImageSize->{Automatic, 25}],
           Appearance->None,
           
           ButtonFunction:>OutputSizeLimit`ButtonFunction[
            Identity, 44, 29881072940564082911, 5/2],
           Enabled->True,
           Evaluator->Automatic,
           Method->"Queued"], 
          ButtonBox[
           PaneSelectorBox[{False->
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource["FEStrings", "sizeShowMore"], 
                StandardForm],
               ImageSizeCache->{56., {0., 8.}}],
              StripOnInput->False,
              DynamicUpdating->True], "SuggestionsBarControl",
             StripOnInput->False], True->
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource["FEStrings", "sizeShowMore"], 
                StandardForm]],
              StripOnInput->False,
              DynamicUpdating->True], "SuggestionsBarControlActive",
             StripOnInput->False]}, Dynamic[
             CurrentValue["MouseOver"]],
            Alignment->Center,
            FrameMargins->0,
            ImageSize->{Automatic, 25}],
           Appearance->None,
           
           ButtonFunction:>OutputSizeLimit`ButtonFunction[
            Identity, 44, 29881072940564082911, 5 2],
           Enabled->True,
           Evaluator->Automatic,
           Method->"Queued"], 
          ButtonBox[
           PaneSelectorBox[{False->
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource["FEStrings", "sizeShowAll"], 
                StandardForm],
               ImageSizeCache->{42., {0., 8.}}],
              StripOnInput->False,
              DynamicUpdating->True], "SuggestionsBarControl",
             StripOnInput->False], True->
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource["FEStrings", "sizeShowAll"], 
                StandardForm]],
              StripOnInput->False,
              DynamicUpdating->True], "SuggestionsBarControlActive",
             StripOnInput->False]}, Dynamic[
             CurrentValue["MouseOver"]],
            Alignment->Center,
            FrameMargins->0,
            ImageSize->{Automatic, 25}],
           Appearance->None,
           
           ButtonFunction:>OutputSizeLimit`ButtonFunction[
            Identity, 44, 29881072940564082911, Infinity],
           Enabled->True,
           Evaluator->Automatic,
           Method->"Queued"], 
          ButtonBox[
           PaneSelectorBox[{False->
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource["FEStrings", "sizeChangeLimit"], 
                StandardForm],
               ImageSizeCache->{73., {0., 8.}}],
              StripOnInput->False,
              DynamicUpdating->True], "SuggestionsBarControl",
             StripOnInput->False], True->
            StyleBox[
             StyleBox[
              DynamicBox[ToBoxes[
                FEPrivate`FrontEndResource["FEStrings", "sizeChangeLimit"], 
                StandardForm]],
              StripOnInput->False,
              DynamicUpdating->True], "SuggestionsBarControlActive",
             StripOnInput->False]}, Dynamic[
             CurrentValue["MouseOver"]],
            Alignment->Center,
            FrameMargins->0,
            ImageSize->{Automatic, 25}],
           Appearance->None,
           ButtonFunction:>FrontEndExecute[{
              FrontEnd`SetOptions[
              FrontEnd`$FrontEnd, 
               FrontEnd`PreferencesSettings -> {"Page" -> "Evaluation"}], 
              FrontEnd`FrontEndToken["PreferencesDialog"]}],
           Evaluator->None,
           Method->"Preemptive"]}
        },
        AutoDelete->False,
        FrameStyle->GrayLevel[0.85],
        GridBoxDividers->{"Columns" -> {False, {True}}},
        GridBoxItemSize->{"Columns" -> {{Automatic}}, "Rows" -> {{Automatic}}},
        GridBoxSpacings->{"Columns" -> {{2}}}]}
     },
     DefaultBaseStyle->"Column",
     GridBoxAlignment->{
      "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
       "RowsIndexed" -> {}},
     GridBoxDividers->{
      "Columns" -> {{False}}, "ColumnsIndexed" -> {}, "Rows" -> {{False}}, 
       "RowsIndexed" -> {}},
     GridBoxItemSize->{
      "Columns" -> {{Automatic}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
       "RowsIndexed" -> {}},
     GridBoxSpacings->{"Columns" -> {
         Offset[0.27999999999999997`], {
          Offset[0.5599999999999999]}, 
         Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
         Offset[0.2], 
         Offset[1.2], {
          Offset[0.4]}, 
         Offset[0.2]}, "RowsIndexed" -> {}}],
    Background->RGBColor[0.9657, 0.9753, 0.9802],
    FrameMargins->{{12, 12}, {0, 15}},
    FrameStyle->GrayLevel[0.85],
    RoundingRadius->5,
    StripOnInput->False],
   Deploy,
   DefaultBaseStyle->"Deploy"],
  Out[44]]], "Output",
 CellChangeTimes->{
  3.6697308478029013`*^9, {3.669731920933063*^9, 3.669731936509972*^9}, 
   3.669970318465561*^9, 3.669970419743229*^9, 3.670073456031206*^9, 
   3.670073647708755*^9, 3.67015054970297*^9, 3.6701507071449003`*^9, 
   3.670410625220437*^9, 3.670410673433688*^9, 3.670410728652534*^9, 
   3.670411551365369*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"reducible", "=", 
  RowBox[{"ExtractReducibleWords", "[", "RedSys", "]"}]}]], "Input",
 CellChangeTimes->{{3.6699702657208242`*^9, 3.669970268124913*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"F", ",", "F"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"D", ",", "F"}], "}"}], ",", 
   RowBox[{"{", "K", "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.669970318579753*^9, 3.66997041983807*^9, 
  3.670073456143289*^9, 3.670073647804701*^9, 3.670150549786736*^9, 
  3.670150707234828*^9, 3.6704106253259373`*^9, 3.670410673613449*^9, 
  3.670410728742585*^9, 3.67041155159271*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Select", "[", 
   RowBox[{"wordsoflength10", ",", 
    RowBox[{
     RowBox[{"!", 
      RowBox[{"ReducibleWordQ", "[", 
       RowBox[{"#", ",", "reducible"}], "]"}]}], "&"}]}], "]"}], "//", 
  "Timing"}]], "Input",
 CellChangeTimes->{{3.669730812903219*^9, 3.669730821171359*^9}, 
   3.66973195790103*^9, {3.66997027255103*^9, 3.669970273907537*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "2.18753500000000000724753590475302189589`6.3605549236300964", ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       OverscriptBox["F", "~"], ",", "D", ",", "D", ",", "D", ",", "D", ",", 
       "D", ",", "D", ",", "D", ",", "D", ",", "D"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{
      "D", ",", "D", ",", "D", ",", "D", ",", "D", ",", "D", ",", "D", ",", 
       "D", ",", "D", ",", "D"}], "}"}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.669730824912211*^9, 3.66973162491469*^9, 
  3.669731974597834*^9, 3.669970320718745*^9, 3.669970422004569*^9, 
  3.670073458352126*^9, 3.6700736498825283`*^9, 3.6701505519170637`*^9, 
  3.670150709420479*^9, 3.670410625453061*^9, 3.670410675742201*^9, 
  3.670410730769988*^9, 3.6704115537994137`*^9}]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
WindowSize->{808, 609},
WindowMargins->{{Automatic, 68}, {44, Automatic}},
FrontEndVersion->"10.0 for Mac OS X x86 (32-bit, 64-bit Kernel) (December 4, \
2014)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "Info3670418748-3520926"->{
  Cell[1147, 36, 269, 5, 43, "Print",
   CellTags->"Info3670418748-3520926"],
  Cell[1635, 52, 244, 4, 43, "Print",
   CellTags->"Info3670418748-3520926"],
  Cell[2044, 65, 281, 5, 43, "Print",
   CellTags->"Info3670418748-3520926"],
  Cell[2486, 79, 287, 5, 61, "Print",
   CellTags->"Info3670418748-3520926"]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"Info3670418748-3520926", 52071, 1542}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[580, 22, 301, 5, 28, "Input"],
Cell[CellGroupData[{
Cell[906, 31, 238, 3, 54, "Print"],
Cell[1147, 36, 269, 5, 43, "Print",
 CellTags->"Info3670418748-3520926"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1465, 47, 167, 3, 28, "Input"],
Cell[1635, 52, 244, 4, 43, "Print",
 CellTags->"Info3670418748-3520926"]
}, Open  ]],
Cell[CellGroupData[{
Cell[1916, 61, 125, 2, 28, "Input"],
Cell[2044, 65, 281, 5, 43, "Print",
 CellTags->"Info3670418748-3520926"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2362, 75, 121, 2, 28, "Input"],
Cell[2486, 79, 287, 5, 61, "Print",
 CellTags->"Info3670418748-3520926"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2810, 89, 126, 1, 44, "Subsection"],
Cell[2939, 92, 143, 4, 28, "Input"],
Cell[3085, 98, 385, 10, 36, "Input"],
Cell[3473, 110, 403, 11, 28, "Input"],
Cell[3879, 123, 256, 7, 28, "Input"],
Cell[4138, 132, 758, 24, 80, "Input"],
Cell[4899, 158, 207, 6, 28, "Input"],
Cell[CellGroupData[{
Cell[5131, 168, 118, 1, 35, "Subsubsection"],
Cell[5252, 171, 1870, 55, 148, "Input"],
Cell[7125, 228, 780, 23, 63, "Input"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7954, 257, 106, 1, 44, "Subsection"],
Cell[8063, 260, 427, 10, 28, "Input"],
Cell[8493, 272, 369, 11, 28, "Input"],
Cell[8865, 285, 240, 7, 28, "Input"],
Cell[9108, 294, 701, 21, 28, "Input"],
Cell[CellGroupData[{
Cell[9834, 319, 120, 1, 35, "Subsubsection"],
Cell[9957, 322, 745, 26, 32, "Text"],
Cell[CellGroupData[{
Cell[10727, 352, 289, 7, 28, "Input"],
Cell[11019, 361, 1061, 24, 28, "Output"]
}, Open  ]],
Cell[12095, 388, 1303, 38, 85, "Text"],
Cell[CellGroupData[{
Cell[13423, 430, 340, 8, 28, "Input"],
Cell[13766, 440, 190, 5, 22, "Print"],
Cell[13959, 447, 460, 15, 28, "Output"]
}, Open  ]],
Cell[14434, 465, 124, 1, 30, "Text"],
Cell[CellGroupData[{
Cell[14583, 470, 336, 8, 28, "Input"],
Cell[CellGroupData[{
Cell[14944, 482, 234, 7, 22, "Print"],
Cell[15181, 491, 234, 7, 22, "Print"],
Cell[15418, 500, 234, 7, 22, "Print"]
}, Open  ]],
Cell[15667, 510, 531, 16, 28, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[16259, 533, 116, 1, 44, "Subsection"],
Cell[CellGroupData[{
Cell[16400, 538, 273, 5, 28, "Input"],
Cell[16676, 545, 1865, 45, 63, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[18578, 595, 277, 8, 28, "Input"],
Cell[18858, 605, 2072, 52, 63, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[20967, 662, 161, 3, 28, "Input"],
Cell[21131, 667, 862, 14, 28, "Output"]
}, Open  ]],
Cell[22008, 684, 134, 1, 30, "Text"],
Cell[CellGroupData[{
Cell[22167, 689, 246, 4, 28, "Input"],
Cell[22416, 695, 868, 14, 28, "Output"]
}, Open  ]],
Cell[23299, 712, 132, 1, 30, "Text"],
Cell[CellGroupData[{
Cell[23456, 717, 204, 4, 28, "Input"],
Cell[CellGroupData[{
Cell[23685, 725, 267, 6, 22, "Print"],
Cell[23955, 733, 328, 8, 22, "Print"],
Cell[24286, 743, 279, 6, 22, "Print"]
}, Open  ]],
Cell[24580, 752, 160, 3, 28, "Output"]
}, Open  ]],
Cell[24755, 758, 140, 1, 30, "Text"],
Cell[CellGroupData[{
Cell[24920, 763, 202, 4, 28, "Input"],
Cell[CellGroupData[{
Cell[25147, 771, 551, 14, 22, "Print"],
Cell[25701, 787, 235, 7, 22, "Print"],
Cell[25939, 796, 235, 7, 22, "Print"],
Cell[26177, 805, 521, 13, 22, "Print"],
Cell[26701, 820, 237, 7, 22, "Print"],
Cell[26941, 829, 235, 7, 22, "Print"],
Cell[27179, 838, 235, 7, 22, "Print"],
Cell[27417, 847, 235, 7, 22, "Print"],
Cell[27655, 856, 237, 7, 22, "Print"],
Cell[27895, 865, 507, 13, 22, "Print"],
Cell[28405, 880, 507, 13, 22, "Print"],
Cell[28915, 895, 481, 13, 22, "Print"],
Cell[29399, 910, 209, 6, 22, "Print"]
}, Open  ]],
Cell[29623, 919, 160, 3, 28, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[29832, 928, 103, 1, 44, "Subsection"],
Cell[29938, 931, 258, 5, 49, "Text"],
Cell[CellGroupData[{
Cell[30221, 940, 579, 18, 36, "Input"],
Cell[30803, 960, 1229, 29, 35, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[32069, 994, 458, 13, 28, "Input"],
Cell[32530, 1009, 996, 26, 35, "Output"]
}, Open  ]],
Cell[33541, 1038, 261, 5, 49, "Text"],
Cell[CellGroupData[{
Cell[33827, 1047, 215, 5, 28, "Input"],
Cell[34045, 1054, 1179, 29, 35, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[35261, 1088, 988, 28, 57, "Input"],
Cell[36252, 1118, 935, 25, 35, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[37224, 1148, 100, 1, 35, "Subsubsection"],
Cell[37327, 1151, 334, 6, 30, "Text"],
Cell[CellGroupData[{
Cell[37686, 1161, 322, 7, 28, "Input"],
Cell[38011, 1170, 799, 18, 35, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[38847, 1193, 583, 16, 36, "Input"],
Cell[39433, 1211, 10355, 259, 183, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[49825, 1475, 178, 3, 28, "Input"],
Cell[50006, 1480, 454, 11, 28, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[50497, 1496, 387, 10, 28, "Input"],
Cell[50887, 1508, 823, 18, 35, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
